from IPython.testing.globalipapp import get_ipython
from IPython.utils.capture import capture_output


class TestSmartPip:
    ip = get_ipython()

    def test_smart_pip(self):
        res = self.ip.run_cell("import smart_pip")
        assert res.success
        self.ip.run_line_magic("pip", "install nothing==0.0.2 -f data")
        self.ip.run_cell("import nothing")
        with capture_output() as captured:
            self.ip.run_line_magic("pip", "install nothing==0.0.3 -f data")
        assert (
            "You should restart Python because the underlying files are updated for these imported modules: nothing"
            in captured.stdout
        )

    def test_smart_pip_when_orignal_pip_nonexistent(self):
        self.ip.run_cell("import smart_pip")
        self.ip.run_line_magic("pip", "install nothing==0.0.2 -f data")
        self.ip.run_cell("import nothing")
        with capture_output() as captured:
            self.ip.run_line_magic("pip", "install nothing==0.0.3 -f data")
        assert (
            "You should restart Python because the underlying files are updated for these imported modules: nothing"
            in captured.stdout
        )

    def test_smart_pip_with_module_reload_hook(self):
        self.ip.run_cell("""
                        import smart_pip
                        def h(s):
                            print('hook called on ' + str(s))
                        smart_pip.add_reload_hook(h)
                         """)
        self.ip.run_line_magic("pip", "install nothing==0.0.2 -f data")
        self.ip.run_cell("import nothing")
        with capture_output() as captured:
            self.ip.run_line_magic("pip", "install nothing==0.0.3 -f data")
        assert (
            "You should restart Python because the underlying files are updated for these imported modules: nothing"
            in captured.stdout
        )
        assert "hook called on ['nothing']" in captured.stdout

    def test_smart_pip_with_all_hooks_cleare(self):
        self.ip.run_cell("""
                        import smart_pip
                        def h(s):
                            print('hook called on ' + str(s))
                        smart_pip.add_reload_hook(h)
                         """)
        self.ip.run_cell("smart_pip.clear_all_hooks()")
        self.ip.run_line_magic("pip", "install nothing==0.0.2 -f data")
        self.ip.run_cell("import nothing")
        with capture_output() as captured:
            self.ip.run_line_magic("pip", "install nothing==0.0.3 -f data")
        assert "hook called on ['nothing']" not in captured.stdout



import subprocess
from IPython import get_ipython
from imports.analyze import (
    _get_dists,
    _build_module_to_dist_map,
    get_imported_modules,
)
from typing import Callable, List


__all__ = ["pip", "add_reload_hook", "clear_all_hooks"]


_module_reload_hooks = []


def add_reload_hook(hook: Callable[[List[str]], None]) -> None:
    """Add a reload hook that acts when modules to reload are detected after pip install. The hook takes the set of modules to relaod as input."""
    _module_reload_hooks.append(hook)


def clear_all_hooks() -> None:
    _module_reload_hooks.clear()


ip = get_ipython()

# Try to preserve the original %pip behavior into run_pip
try:
    run_pip = ip.magics_manager.magics["line"]["pip"]
except KeyError:
    # If %pip is not ever registered, create a default one using subprocess
    def run_pip(line):
        ret = subprocess.run(["pip"] + line.split(" "))
        if ret.returncode:
            print(
                f"Failed to run pip {line}. stdout: {ret.stdout} | stderr: {ret.stderr}"
            )
            return


BLUE = "\033[94m"
RESET = "\033[0m"


def pip(line):
    """Jupyter line magic %smart_pip"""

    dists_snapshot = _get_dists()

    run_pip(line)

    added_dists = _get_dists() - dists_snapshot
    if added_dists:
        affected_modules = _build_module_to_dist_map(added_dists).keys()
        imported_modules = set(get_imported_modules())
        modules_to_reload = list(imported_modules & affected_modules)
        if modules_to_reload:
            print(
                f'{BLUE}Note: You should restart Python because the underlying files are updated for these imported modules: {", ".join(modules_to_reload)}{RESET}'
            )
            for h in _module_reload_hooks:
                h(modules_to_reload)


# Register the smart %pip line magic and override the orignal one
ip.magics_manager.register_function(pip)



import importlib
import sys


from typing import List, Dict, Tuple, Set

__all__ = ["get_imported_modules", "get_imported_3rd_party_modules"]


def get_imported_modules() -> List[str]:
    """Returns all imported modules."""
    return list(sys.modules.keys())


def get_imported_3rd_party_modules() -> Dict[str, Tuple[str, str]]:
    """Returns all 3rd-party imported modules and their associated (dist, version)"""
    result = {}
    module_to_dist = _build_module_to_dist_map()
    for mod in get_imported_modules():
        if mod in module_to_dist:
            dist, version = module_to_dist[mod]
            result.update({mod: (dist, version)})
    return result


def _get_dists() -> Set[str]:
    """Retruns a snapshot of the current available dist in form of '[name]-[version]'"""
    return set(
        map(
            lambda dist: f"{dist.name}-{dist.version}",
            importlib.metadata.distributions(),
        )
    )


def _build_module_to_dist_map(
    dists: Set[str] = None,
) -> Dict[str, Tuple[str, str]]:
    """Returns a dict mapping module names to a tuple of (dist name, version).

    dists -- the given dicts of the same format from _get_dists(). If not specified, we fetch from _get_dists()
    """
    module_to_dep = {}
    for dist_and_version in dists or _get_dists():
        dist, _, version = dist_and_version.rpartition("-")
        for file_path in importlib.metadata.Distribution.from_name(dist).files:
            file_path = str(file_path)
            mod = None
            if file_path.startswith(".."):
                continue
            if file_path.endswith("/__init__.py"):
                mod = file_path.removesuffix("/__init__.py").replace("/", ".")
            elif file_path.endswith(".py"):
                mod = file_path.removesuffix(".py").replace("/", ".")
            elif file_path.endswith(".so"):
                mod = file_path.partition(".")[0]
            else:
                continue
            module_to_dep.update({mod: (dist, version)})
    return module_to_dep
